package com.smppw.analysis.application.service.info;

import cn.hutool.core.bean.BeanUtil;
import cn.hutool.core.collection.CollUtil;
import cn.hutool.core.collection.ListUtil;
import cn.hutool.core.date.DateUtil;
import cn.hutool.core.map.MapUtil;
import cn.hutool.core.text.CharSequenceUtil;
import cn.hutool.core.util.NumberUtil;
import cn.hutool.core.util.StrUtil;
import com.smppw.analysis.application.dto.info.*;
import com.smppw.analysis.domain.dataobject.*;
import com.smppw.analysis.domain.dto.info.*;
import com.smppw.analysis.domain.dto.performance.IndicatorParams;
import com.smppw.analysis.domain.manager.performance.Performance;
import com.smppw.analysis.domain.manager.performance.PerformanceConstants;
import com.smppw.analysis.domain.manager.performance.PerformanceFactory;
import com.smppw.analysis.domain.service.BaseIndicatorServiceV2;
import com.smppw.analysis.domain.service.BaseInfoService;
import com.smppw.analysis.domain.service.NavService;
import com.smppw.analysis.infrastructure.utils.StringUtil;
import com.smppw.common.pojo.IStrategy;
import com.smppw.common.pojo.NewSubStrategy;
import com.smppw.common.pojo.ValueLabelVO;
import com.smppw.common.pojo.dto.DateValue;
import com.smppw.common.pojo.dto.calc.IndicatorCalcPropertyDto;
import com.smppw.common.pojo.dto.calc.IndicatorCalcTimeRangeDto;
import com.smppw.common.pojo.enums.*;
import com.smppw.constants.DateConst;
import com.smppw.core.IndicatorService;
import com.smppw.utils.BigDecimalUtils;
import com.smppw.utils.CalcUtils;
import com.smppw.utils.NavUtil;
import com.smppw.utils.StrategyHandleUtils;
import org.apache.commons.lang3.time.DateFormatUtils;
import org.springframework.stereotype.Service;

import java.math.RoundingMode;
import java.util.*;
import java.util.function.Function;
import java.util.stream.Collectors;

@Service
public class FundInfoService {
    private final NavService navService;
    private final PerformanceFactory factory;
    private final BaseInfoService baseInfoService;
    private final BaseIndicatorServiceV2 baseIndicatorServiceV2;
    private static final Map<String, String> ACCRUED_FREQUENCY = MapUtil.newHashMap(5);
    private static final Map<String, String> ACCRUED_METHOD = MapUtil.newHashMap(3);
    private static final Map<Integer, String> EDUCATION_NAME_MAP = MapUtil.newHashMap(8);

    public FundInfoService(NavService navService, PerformanceFactory factory,
                           BaseInfoService baseInfoService, BaseIndicatorServiceV2 baseIndicatorServiceV2) {
        this.factory = factory;
        this.navService = navService;
        this.baseInfoService = baseInfoService;
        this.baseIndicatorServiceV2 = baseIndicatorServiceV2;
    }

    static {
        ACCRUED_FREQUENCY.put("1", "月");
        ACCRUED_FREQUENCY.put("2", "季");
        ACCRUED_FREQUENCY.put("3", "半年");
        ACCRUED_FREQUENCY.put("4", "年");
        ACCRUED_FREQUENCY.put("0", "不固定");
        ACCRUED_METHOD.put("1", "高水位");
        ACCRUED_METHOD.put("2", "高水位+赎回时补充");
        ACCRUED_METHOD.put("3", "单客户高水位");
        EDUCATION_NAME_MAP.put(1,"小学");
        EDUCATION_NAME_MAP.put(2,"中学");
        EDUCATION_NAME_MAP.put(3,"大专");
        EDUCATION_NAME_MAP.put(4,"本科");
        EDUCATION_NAME_MAP.put(5,"硕士");
        EDUCATION_NAME_MAP.put(6,"博士");
        EDUCATION_NAME_MAP.put(7,"博士后");
        EDUCATION_NAME_MAP.put(-1,"其他");
    }

    public PrivatelyFundHeadInfoVO headInfo(HeadInfoReq params) {
        return new PrivatelyFundHeadInfoVO();
    }

    public CommonInfoVO commonInfo(CommonInfoReq params) {
//        CommonInfoVO res = new CommonInfoVO();
//        IStrategy strategy = StrategyHandleUtils.getStrategy(params.getStrategy());
//        String secId = params.getSecId();
//        Byte raiseType = null;
//        List<Frequency> result = CollectionUtil.newArrayList();
//        String fundType = this.baseInfoService.getFundType(secId);
//        List<IStrategy> originStrategy = CollectionUtil.newArrayList();
//        // 获取频率
//        CmBigDataFundInfoDo fundHeadInfo = fundHeadInfoBaseService.getFundHeadInfo(secId);
//        if (fundHeadInfo == null) {
//            throw new APIException("");
//        }
//        raiseType = fundHeadInfo.getRaiseType();
//        Frequency navFrequency = fundInformationDao.getNavFrequency(secId);
//        result.add(navFrequency);
//        // 将频率转换一道 取最小频率
//        List<Frequency> frequencies = fundInformationDao.toFrequencyList(result);
//        if (raiseType == 2) {
//            frequencies.removeIf(p -> p.getId() == Frequency.Default.getId());
//        }
//        List<ValueLabelVO> frequencyValueLabel = trans2ValueLabel(frequencies);
//        res.setFrequency(frequencyValueLabel);
//        // 净值类型
//        List<ValueLabelVO> navType = Arrays.stream(NavType.values())
//                .filter(p -> !NavType.All.equals(p) && !NavType.UnitAndCumulativeNav.equals(p))
//                .map(p -> new ValueLabelVO(p.name(), p.getDesc())).collect(Collectors.toList());
//        res.setNavType(navType);
//        // 策略曲线
//        RaiseType realRaiseType = RaiseType.getRaiseType(raiseType);
//        handleTimeRange(strategy, res, secId, companyOrManager, realRaiseType);
//        // 获取常用指数
////        List<ValueLabelVO> benchmarkList = userCommonIndexService.getByUser(-1);
////        handleBenchmark(strategy, res, secId, fundType, originStrategy, companyOrManager, benchmarkList);
        return new CommonInfoVO();
    }

    @SuppressWarnings("unchecked")
    public HeadIndicatorVO headIndicator(HeadIndicatorReq req) {
        IndicatorParams params = req.convert();
        Performance<IndicatorParams, Map<String, Object>> instance = this.factory.getInstance(PerformanceConstants.INDICATOR);
        Map<String, Object> dataset = instance.execute(params);
        Map<String, Object> indicatorMap = MapUtil.get(dataset, req.getSecId(), Map.class);
        HeadIndicatorVO vo = new HeadIndicatorVO();
        vo.setAnnualReturn(MapUtil.getStr(indicatorMap, Indicator.AnnualReturn.name()));
        vo.setAnnualStdDev(MapUtil.getStr(indicatorMap, Indicator.AnnualStdDev.name()));
        vo.setMaxDrawdown(MapUtil.getStr(indicatorMap, Indicator.MaxDrawdown.name()));
        vo.setSharpeRatio(MapUtil.getStr(indicatorMap, Indicator.SharpeRatio.name()));
        return vo;
    }

    /**
     * 私募基金相似产品，计算相关性时代码可以优化（参考基金相关性）
     *
     * @param req /
     * @return /
     */
    public List<FundSimilarVO> getFundSimilarList(FundSimilarReq req) {
        List<String> refIds = ListUtil.toLinkedList(req.getRefId());
        IStrategy strategy = StrategyHandleUtils.getStrategy(req.getStrategy());
        Map<String, List<DateValue>> secNavMap = this.navService.getSecIdDateValueNavListMapByDb(refIds,
                null, null, Visibility.Both, NavType.CumulativeNav, MapUtil.empty());
        List<IndicatorCalcTimeRangeDto> secTimeRanges = IndicatorService.getInstance().getSecTimeRange(req.getRefId(), Frequency.Monthly, secNavMap);
        String startDate = null;
        String endDate = null;
        if (CollUtil.isNotEmpty(secTimeRanges)) {
            for (IndicatorCalcTimeRangeDto secTimeRange : secTimeRanges) {
                if (TimeRange.Last1Year == secTimeRange.getTimeRange()) {
                    startDate = secTimeRange.getStartDate();
                    endDate = secTimeRange.getEndDate();
                    break;
                }
            }
        }
        FundSimilarParams params = req.convert();
        params.setStartDate(startDate);
        List<FundSimilarDo> dataList = this.baseInfoService.getFundSimilarList(params);
        List<String> fundIds = dataList.stream().map(FundSimilarDo::getFundId).collect(Collectors.toList());
        if (CollUtil.isNotEmpty(fundIds)) {
            refIds.addAll(fundIds);
        }
        Map<String, List<DateValue>> allNavMap = this.navService.getSecIdDateValueNavListMapByDb(refIds,
                startDate, endDate, Visibility.Both, NavType.CumulativeNav, MapUtil.empty());
        Map<String, List<IndicatorCalcPropertyDto>> multiSecRetListNew = this.baseIndicatorServiceV2.getMultiSecRetListNew(
                refIds, ListUtil.empty(), Frequency.Monthly, Frequency.Daily, startDate, endDate, false, params.getBenchmarkId(),
                params.getRaiseType(), strategy, Visibility.Both, NavType.CumulativeNav, allNavMap);
        List<Map<String, Object>> initRetList = NavUtil.handleNextStepNew(refIds, ListUtil.empty(), multiSecRetListNew, false);
        Map<String, Map<String, Double>> retMap = this.transfer2TwoDimensionMap(initRetList);
        Double[][] corArr = CalcUtils.getCorrelationMatrix(retMap, ListUtil.toList(params.getRefId()), refIds);
        List<FundSimilarVO> resultList = ListUtil.list(true);
        for (FundSimilarDo map : dataList) {
            FundSimilarVO similarVO = BeanUtil.copyProperties(map, FundSimilarVO.class);
            // 因为第一个基金是当前基金，所以需要 + 1 来取该基金与当前基金的相关性
            int i = fundIds.indexOf(map.getFundId()) + 1;
            similarVO.setSimilar(corArr[0][i]);
            similarVO.setStrategy(StrategyHandleUtils.getStrategy(map.getStrategy()).getStrategyNameDesc());
            similarVO.setSubStrategy(this.getStrategyName(map.getSubStrategy()));
            similarVO.setThirdStrategy(this.getStrategyName(map.getThirdStrategy()));
            resultList.add(similarVO);
        }
        return resultList.stream().filter(e -> e.getSimilar() != null && e.getSimilar() >= Double.parseDouble(params.getThreshold())).collect(Collectors.toList());
    }

    private String getStrategyName(String strategyKey) {
        IStrategy strategy = StrategyHandleUtils.getStrategy(strategyKey);
        if (strategy == null || strategy == NewSubStrategy.OtherStrategy || strategy.getStrategyId() == -1) {
            return "--";
        }
        return strategy.getStrategyNameDesc();
    }

    private Map<String, Map<String, Double>> transfer2TwoDimensionMap(List<Map<String, Object>> initRetList) {
        Map<String, Map<String, Double>> map = new TreeMap<>();
        for (Map<String, Object> unit : initRetList) {
            if (unit != null) {
                String id = unit.get("fund_id").toString();
                if (!map.containsKey(id)) {
                    map.put(id, new TreeMap<>());
                }
                Object t = unit.get("ret_range");
                if (t == null) {
                    continue;
                }
                map.get(id).put(unit.get("end_date").toString(), Double.valueOf(unit.get("ret_range").toString()));
            }

        }
        return map;
    }

    public List<ManualFundNoticeInfoVO> noticeList(ManualFundNoticeReq params) {
        Map<String, String> typeMapper = MapUtil.builder("1", "10").put("2", "20").put("3", "30").put("4", "50, 70, 90, 91").build();
        String fundId = params.getRefId();
        String title = params.getTitle();
        // type 逗号拆分 转list，直接数据库in查询
        List<String> typeList = ListUtil.list(false);
        if (CharSequenceUtil.isNotBlank(params.getType())) {
            String s = typeMapper.get(params.getType());
            typeList.addAll(StrUtil.split(s, ","));
        }
        //todo 确认基金公告表所在的库
        List<ManualFundNoticeInfoDO> dataList = this.baseInfoService.queryFundAnnounce(fundId, typeList, title);
        List<ManualFundNoticeInfoVO> resultList = ListUtil.list(true);
        // key value互换
        Map<String, String> reverse = MapUtil.reverse(typeMapper);
        for (ManualFundNoticeInfoDO temp : dataList) {
            ManualFundNoticeInfoVO vo = BeanUtil.copyProperties(temp, ManualFundNoticeInfoVO.class);
            vo.setType(reverse.getOrDefault(vo.getType(), "4"));
            vo.setDate(DateUtil.formatDate(temp.getDate()));
            resultList.add(vo);
        }
        return resultList;
    }

    public ManualFundInvestInfoVO investInfo(String refId) {
        ManualFundInvestInfoVO fundInvestInfoVO = new ManualFundInvestInfoVO();
        FundArchivesInfoDO fundArchivesInfo = this.baseInfoService.queryFundArchives(refId);
        if (fundArchivesInfo == null) {
            return fundInvestInfoVO;
        }
        fundInvestInfoVO.setInvestmentObjective(fundArchivesInfo.getInvestmentObjective());
        fundInvestInfoVO.setStrategyDescription(fundArchivesInfo.getStrategyDescription());
        fundInvestInfoVO.setInvestmentScope(fundArchivesInfo.getInvestmentScope());
        return fundInvestInfoVO;
    }

    public FundBaseFeeVO getFundFee(String refId) {
        FundFeeDo fundFee = this.baseInfoService.getFundFee(refId);
        DateFormatUtils.format(new Date(), DateConst.YYYY, TimeZone.getDefault());
        //数字格式处理
        FundFeeDo fundFees = numberFormat(fundFee);
        FundBaseFeeVO fee = BeanUtil.copyProperties(fundFees, FundBaseFeeVO.class);
        String frequency = Optional.ofNullable(fundFees.getAccruedFrequency()).map(e -> StrUtil.subBetween(e, "[", "]"))
                .map(e -> StrUtil.split(e, ",")).orElse(ListUtil.empty()).stream().filter(StrUtil::isNotBlank)
                .map(ACCRUED_FREQUENCY::get).collect(Collectors.joining(","));
        fee.setAccruedFrequency(frequency);
        String method = Optional.ofNullable(fundFees.getAccruedMethod()).map(e -> StrUtil.subBetween(e, "[", "]"))
                .map(e -> StrUtil.split(e, ",")).orElse(ListUtil.empty()).stream().filter(StrUtil::isNotBlank)
                .map(ACCRUED_METHOD::get).collect(Collectors.joining(","));
        fee.setAccruedMethod(method);
        return fee;
    }

    private FundFeeDo numberFormat(FundFeeDo fundFees) {
        FundFeeDo fundFeeDo = fundFees;
        String minInvestmentShare = fundFees.getMinInvestmentShare();
        fundFeeDo.setMinInvestmentShare(minInvestmentShare != null && StringUtil.isNumeric(minInvestmentShare) ? BigDecimalUtils.toBigDecimal(minInvestmentShare)
                .setScale(2, RoundingMode.HALF_DOWN) + "万" : minInvestmentShare);

        String managementFeeBank = fundFees.getManagementFeeBank();
        fundFeeDo.setManagementFeeBank(managementFeeBank != null && StringUtil.isNumeric(managementFeeBank) ? BigDecimalUtils.toBigDecimal(managementFeeBank)
                .setScale(2, RoundingMode.HALF_DOWN) + "%" : managementFeeBank);

        String subscriptionFee = fundFees.getSubscriptionFee();
        fundFeeDo.setSubscriptionFee(subscriptionFee != null && StringUtil.isNumeric(subscriptionFee) ? BigDecimalUtils.toBigDecimal(subscriptionFee)
                .setScale(2, RoundingMode.HALF_DOWN) + "%" : subscriptionFee);

        String redemptionFee = fundFees.getRedemptionFee();
        fundFeeDo.setRedemptionFee(redemptionFee != null && StringUtil.isNumeric(redemptionFee) ? BigDecimalUtils.toBigDecimal(redemptionFee)
                .setScale(2, RoundingMode.HALF_DOWN) + "%" : redemptionFee);

        String managementFee = fundFees.getManagementFee();
        fundFeeDo.setManagementFee(managementFee != null && StringUtil.isNumeric(managementFee) ? BigDecimalUtils.toBigDecimal(managementFee)
                .setScale(2, RoundingMode.HALF_DOWN) + "%" : managementFee);

        String managementFeeConsult = fundFees.getManagementFeeConsult();
        fundFeeDo.setManagementFeeConsult(managementFeeConsult != null && StringUtil.isNumeric(managementFeeConsult) ? BigDecimalUtils.toBigDecimal(managementFeeConsult)
                .setScale(2, RoundingMode.HALF_DOWN) + "%" : managementFeeConsult);

        String performanceFee = fundFees.getPerformanceFee();
        fundFeeDo.setPerformanceFee(performanceFee != null && StringUtil.isNumeric(performanceFee) ? BigDecimalUtils.toBigDecimal(performanceFee)
                .setScale(2, RoundingMode.HALF_DOWN) + "%" : performanceFee);

        String guardLine = fundFees.getGuardLine();
        fundFeeDo.setGuardLine(guardLine != null && StringUtil.isNumeric(guardLine) ? BigDecimalUtils.toBigDecimal(guardLine)
                .setScale(2, RoundingMode.HALF_DOWN) + "%" : guardLine);

        String subsequentInvestmentShare = fundFees.getSubsequentInvestmentShare();
        fundFeeDo.setSubsequentInvestmentShare(subsequentInvestmentShare != null && StringUtil.isNumeric(subsequentInvestmentShare) ? BigDecimalUtils.toBigDecimal(subsequentInvestmentShare)
                .setScale(2, RoundingMode.HALF_DOWN) + "万" : subsequentInvestmentShare);

        String stopLossLine = fundFees.getStopLossLine();
        fundFeeDo.setStopLossLine(stopLossLine != null && StringUtil.isNumeric(stopLossLine) ? BigDecimalUtils.toBigDecimal(stopLossLine)
                .setScale(2, RoundingMode.HALF_DOWN) + "%" : stopLossLine);
        return fundFeeDo;
    }

    public ManualFundFeeInfoVO getMfFundFee(String fundId) {
        List<ManualFundFeeDo> dataList = this.baseInfoService.getManualFundFee(fundId);
        ManualFundFeeInfoVO result = new ManualFundFeeInfoVO();
        // 数据库 费率 类型与中文映射
        Map<String, String> typeMapper = MapUtil.builder("10210", "认购金额").put("11210", "申购金额").put("11010", "申购金额")
                .put("12200", "持有期限").put("12000", "持有期限").build();
        // 费率类型 与 单位映射
        Map<String, String> unitMapper = MapUtil.builder("10210", "万元").put("11210", "万元").put("11010", "万元")
                .put("12200", "日").put("12000", "日").build();
        // 记录映射函数，返回一个包含最大最小的范围
        Function<ManualFundFeeDo, ValueLabelVO> function = e -> {
            ValueLabelVO vo = new ValueLabelVO();
            vo.setValue(e.getChargeRateDes());
            String label = typeMapper.get(e.getChargeRateType());
            // 最小值不为0和null时，设置为起始数据
            if (e.getStDivStand1() != null && !Double.valueOf(0.0d).equals(e.getStDivStand1())) {
                label = NumberUtil.roundStr(e.getStDivStand1(), 0) + unitMapper.get(e.getChargeRateType()) + " <= " + label;
            }
            // 最大值不为null时设置截止范围数据
            if (e.getEnDivStand1() != null) {
                label += " < " + NumberUtil.roundStr(e.getEnDivStand1(), 0) + unitMapper.get(e.getChargeRateType());
            }
            vo.setLabel(label);
            return vo;
        };
        // 认购费
        List<ValueLabelVO> subscriptionFee = this.handleFee(dataList, "10210", function);
        result.setSubscriptionFee(subscriptionFee);
        // 申购费，先取场外
        List<ValueLabelVO> applyFee = this.handleFee(dataList, "11210", function);
        if (applyFee.isEmpty()) {
            applyFee = this.handleFee(dataList, "11010", function);
        }
        result.setApplyFee(applyFee);
        // 赎回费，先取场外
        List<ValueLabelVO> redemptionFee = this.handleFee(dataList, "12200", function);
        if (redemptionFee.isEmpty()) {
            redemptionFee = this.handleFee(dataList, "12000", function);
        }
        result.setRedemptionFee(redemptionFee);
        // 管理费、托管费和营销费
        result.setManagementFeeTrust(this.handleOneFee(dataList, "15000", ManualFundFeeDo::getChargeRateDes));
        result.setManagementFeeBank(this.handleOneFee(dataList, "16000", ManualFundFeeDo::getChargeRateDes));
        result.setSaleFee(this.handleOneFee(dataList, "19000", ManualFundFeeDo::getChargeRateDes));
        return result;
    }

    /**
     * 费率处理工具，返回一个列表
     *
     * @param dataList 费率数据
     * @param type     过滤的关键字
     * @param function 转换函数
     * @param <T>      类型参数
     * @return /
     */
    private <T> List<T> handleFee(List<ManualFundFeeDo> dataList, String type, Function<ManualFundFeeDo, T> function) {
        return dataList.stream().filter(e -> type.equals(e.getChargeRateType())).map(function).collect(Collectors.toList());
    }

    /**
     * 费率处理工具，返回一个对象
     *
     * @param dataList 费率数据
     * @param type     过滤的关键字
     * @param function 转换函数
     * @param <T>      类型参数
     * @return /
     */
    private <T> T handleOneFee(List<ManualFundFeeDo> dataList, String type, Function<ManualFundFeeDo, T> function) {
        return this.handleFee(dataList, type, function).stream().findFirst().orElse(null);
    }

    public List<FundManagerInfoVo> getHFManagerInfo(String refId) {
        List<FundManagerInfoVo> resultList = ListUtil.list(true);
        List<PersonnelInformationDo> managerInfoList = this.baseInfoService.getFundManagerInfo(refId);
        if (CollUtil.isEmpty(managerInfoList)) {
            return resultList;
        }
        List<String> managerIdList = managerInfoList.stream().map(PersonnelInformationDo::getPersonnelId).collect(Collectors.toList());
        List<PersonnelWorkExperienceDo> workExperienceList = this.baseInfoService.listPersonnelWorkExperience(managerIdList);
        //过滤掉任职开始日期和任职结束日期为空的数据
        if (CollUtil.isNotEmpty(workExperienceList)) {
            workExperienceList = workExperienceList.stream()
                    .filter(e -> !(StrUtil.isEmpty(e.getStartDate()) && StrUtil.isEmpty(e.getEndDate()))).collect(Collectors.toList());
        }
        if (CollUtil.isNotEmpty(workExperienceList)) {
            Map<String, List<PersonnelWorkExperienceDo>> personnelIdWorkExperienceMap = workExperienceList.stream()
                    .collect(Collectors.groupingBy(PersonnelWorkExperienceDo::getPersonnelId));
            for (PersonnelInformationDo fundManagerInfo : managerInfoList) {
                FundManagerInfoVo result = new FundManagerInfoVo();
                result.setManagerId(fundManagerInfo.getPersonnelId());
                result.setManagerName(fundManagerInfo.getPersonnelName());
                result.setAvatar(fundManagerInfo.getAvatar());
                result.setAvatar2(fundManagerInfo.getAvatar2());
                result.setEducation(EDUCATION_NAME_MAP.get(fundManagerInfo.getEducation()));
                result.setProfile(fundManagerInfo.getProfile());
                List<PersonnelWorkExperienceDo> workExperienceDoList = personnelIdWorkExperienceMap.get(fundManagerInfo.getPersonnelId());
                if (CollUtil.isNotEmpty(workExperienceDoList)) {
                    String latestPosition = workExperienceDoList.stream().filter(e -> e.getEndDate() == null)
                            .sorted(Comparator.comparing(PersonnelWorkExperienceDo::getStartDate).reversed())
                            .map(PersonnelWorkExperienceDo::getPosition).findFirst().orElse(null);
                    result.setManagerWorkExperienceList(workExperienceDoList);
                    result.setLatestPosition(latestPosition);
                }
                resultList.add(result);
            }
        }
        return resultList;
    }

}
